####################################################################
# Distributor.rb
# Defines Distributor class(es) which is(are) used to distribute runnable
# task objects (objects which respond to a 'run' message) to multiple 
# clients.
# 12/12/2001
# Copyright (C) 2001 Phil Tomson
# This file covered by the same license as Ruby's
# See 'COPYING' file
#
####################################################################
module TaskMaster 
require "common.rb"
##########################################################################
# DistributorGeneric - manages a collection of SlaveProxy objects
# sends tasks to next free machine
##########################################################################
  class DistributorGeneric
     attr_writer :reporter, :checkUnreachableInterval
     def initialize(clientList,attribsToCheck=nil,reporterObj=nil, sampletime = SAMPLETIME)
        @machineList = clientList
        @sampletime = sampletime
        @unreachableClients = Array.new
        @clientHash = Hash.new
        ##################################################################
        #@checkUnreachableInterval - 
        #if it's nil (default) then the clients listed
        #in @unreachableClients are never checked.  If it's a number then
        #they (the unreachable clients are checked every <number> of seconds 
        #during idle time in 
        #send_task as it waits for a client to become available
        @checkUnreachableInterval = nil
        @counter = 0
        ##################################################################
        #the reporterObj is the object responsible for creating
        #a report from what is 'harvested' from the client
        #must respond to 'report' message
        @reporter = reporterObj
        #start DRb service:
        DRb.start_service()
        clientObjRef = nil
        clientList.each { |client|
           add_client(client,attribsToCheck)
        }
        puts "@clientHash.keys = #{@clientHash.keys}"
     end #initialize

     def add_client(clientStr,attribsToCheck=nil)
     
       begin
         clientObjRef = SlaveProxy.new(clientStr)
         clientObjRef.check_attributes(attribsToCheck) 
       rescue AttributeMismatch => errStr
         puts errStr
         unless timed_question("Is this OK?\n   ( 'n' means don't use this client)")
            return false
         end
       rescue
         puts "#{clientStr} is not currently avaliable!"
         dbg_puts "exception was: #{$!}"
         puts "exception was: #{$!}"
         @unreachableClients << clientStr unless @unreachableClients.include?(clientStr)
         return false #not added
       else
       end
       puts "#{clientStr} is available."
       @clientHash[clientStr] = clientObjRef
       puts "clientObjRef is: #{clientObjRef.to_s}"
       return true
     end

     # availableSlaves - returns a list of currently reachable slaves
     def availableSlaves
       @clientHash.keys
     end


     ###################################################################
     #broadcast - send the same runnable object to all clients
     #useful for initialization of clients
     ###################################################################
     def broadcast(runnableObj,*args)
        @clientHash.each { |clientName,clientObj|
           clientObj.run(runnableObj,args)
        } 
	#the next line is very important:
	self.join_em #wait for all the sent tasks to complete before harvesting!
     end
     ###################################################################
     #add_method2Clients - send a string which defines a method to all
     #clients for instance_eval.  This allows us to add more methods 
     #to clients without stopping them.
     ###################################################################
     def add_method2Clients(methodAsStr)
        @clientHash.each { |clientName,clientObj|
           clientObj.add_method(methodAsStr)
        }
     end
     ###################################################################
     # send_task to the next client which is in the AVAILABLE state
     ###################################################################
     def send_task(runnableObj,*args)
        result = ""
        clientList = @clientHash.keys
        raise ClientListEmpty, "clientList is empty!" if clientList.empty?
        puts "clientList is: #{clientList}" if $DEBUG
        loop {
           clientName = clientList.rotate!
           puts "try: #{clientName}" if $DEBUG
           client = @clientHash[clientName]
           #wait for machine to become available
           puts "#{clientName}'s status is: #{client.status}" if $DEBUG
           case client.status
           when "FINISHED"
              puts "reporter is: #{@reporter.type}"
              client.harvest(@reporter)
              #harvesting will make it AVAILABLE in the next round
           when "AVAILABLE"
              #if the runnableObj has a get_serverTimeLimit method defined
              #call it and assign the value to the client's serverTimeLimit
              if(runnableObj.respond_to?(:get_serverTimeLimit))
                 client.serverTimeLimit = runnableObj.get_serverTimeLimit
              end
              client.run(runnableObj,args)
              #was:return result
              return client #return the client object
           when "TIMED_OUT"
              warnstr = clientName+" has timed out! - removing from clientHash"
              puts warnstr
	      STDERR << warnstr+"\n"
              result = "TIMED_OUT"
              @unreachableClients << clientName 
              @clientHash.delete(clientName)
              client = nil #make sure there is no ref to that client
              if @clientHash.empty?
                 raise "No clients are currently reachable!"
              end
	      #TODO: figure out if this return should be here!
	      #return result
           when "BUSY"
              puts "#{clientName} is BUSY" if $DEBUG 
              #otherwise do nothing
           when "UNKNOWN"
            
           end
           sleep @sampletime
	   if @checkUnreachableInterval
	     @counter += 1
	     if @counter >= @checkUnreachableInterval.to_f/@sampletime.to_f
	       puts "now check_unreachable_clients" if $DEBUG
	       check_unreachable_clients
	       @counter = 0
	       #reload clientList in case clients were added:
               clientList = @clientHash.keys
               puts "Now clientList is: #{clientList}" if $DEBUG
	     end
           end
        }
        return result
     end #send_task

     def check_unreachable_clients
       addedClients = []
       @unreachableClients.each { |clientName|
         addedClients << clientName if add_client(clientName) 
       }
       #addedClients are not longer unreachable:
       @unreachableClients -= addedClients
     end

     ############################################################
     # get_status for a particular client
     ############################################################
     def get_status(clientName)
        return @clientHash[clientName].status
     end


########################################################################
# join_em - this waits for all of the client threads to finish before
# exiting.  It should be called after the TaskMaster object is finished
# or when you want all current tasks to complete before going on
# to distribute more tasks
########################################################################
     def join_em
       puts "TaskMaster::join_em" if $DEBUG
       @clientHash.each {|key,client|
          if client
             client.join_thread 
             if client.status == "FINISHED"
                client.harvest(@reporter)
             end
          end
       }
     end #join_em

    # wait is just a synonym for join_em
     def wait
       join_em
     end

  ########################################################
  # remote_require
  # given a list of filenames sends each file as a string
  # to each client calling the clientObjs (Client's) 
  # remote_require on each of the strings.
  # ######################################################
     def remote_require(fileList)
        fileList.each { |file|
	  #was:fileStr = File.new(file).stringify
	  fileStr = File.open(file){ |f| f.readlines.join }
           dbg_puts "fileStr.type = #{fileStr.type}"
           @clientHash.each { |clientName,clientObj|
              clientObj.remote_require(fileStr)
           }
        }
     end
  end #class Distributor

  class Distributor < DistributorGeneric
  end

end #module TaskMaster

